from compas_fab.backends.ros.messages import ROSmsg

from compas_rrc.common import ExecutionLevel
from compas_rrc.common import FeedbackLevel

INSTRUCTION_PREFIX = 'r_RRC_'

__all__ = [
    'Zone',
    'Motion',
    'MoveToJoints',
    'MoveToFrame',
    'MoveToRobtarget',
]


class Zone(object):
    """Describes the valid zone data definitions.

    There are two types of zones: fine and fly-by.
    If zone is ``FINE``, the movement terminates as a stop point, and the program execution
    will not continue until robot reach the stop point. For all other zones, the movement
    terminates as a fly-by point, and the program execution continues about 100 ms before
    the robot reaches the zone.

    .. autoattribute:: FINE
    .. autoattribute:: Z0
    .. autoattribute:: Z1
    .. autoattribute:: Z5
    .. autoattribute:: Z10
    .. autoattribute:: Z15
    .. autoattribute:: Z20
    .. autoattribute:: Z30
    .. autoattribute:: Z40
    .. autoattribute:: Z50
    .. autoattribute:: Z60
    .. autoattribute:: Z80
    .. autoattribute:: Z100
    .. autoattribute:: Z150
    .. autoattribute:: Z200

    """

    FINE = -1
    """
    Fine point.
    """

    Z0 = 0
    """
    0.3 mm.
    """

    Z1 = 1
    """
    1 mm.
    """

    Z5 = 5
    """
    5 mm.
    """

    Z10 = 10
    """
    10 mm.
    """

    Z15 = 15
    """
    15 mm.
    """

    Z20 = 20
    """
    20 mm.
    """

    Z30 = 30
    """
    30 mm.
    """

    Z40 = 40
    """
    40 mm.
    """

    Z50 = 50
    """
    50 mm.
    """

    Z60 = 60
    """
    60 mm.
    """

    Z80 = 80
    """
    80 mm.
    """

    Z100 = 100
    """
    100 mm.
    """

    Z150 = 150
    """
    150 mm.
    """

    Z200 = 200
    """
    200 mm.
    """


class Motion(object):
    """Represents valid motion types.

    .. autoattribute:: LINEAR
    .. autoattribute:: JOINT
    """
    LINEAR = 'L'
    """Moves the robot linearly to the specified position."""

    JOINT = 'J'
    """Moves the robot not linearly to the specified position by coordinating all joints to start and end together.
    This type of motion can be faster than LINEAR motion."""


class MoveToJoints(ROSmsg):
    """Move to joints is a call that moves the robot and the external axes with axes values.

    Examples
    --------
    .. code-block:: python

        # Get joints
        robot_joints, external_axes = abb.send_and_wait(rrc.GetJoints())

        # Print received values
        print(robot_joints, external_axes)

        # Change value and move to new position
        robot_joints.rax_1 += 15
        speed = 100 # Unit [mm/s]
        done = abb.send_and_wait(rrc.MoveToJoints(robot_joints, external_axes, speed, rrc.Zone.FINE))

    RAPID Instruction: ``MoveAbsJ``

    .. include:: ../abb-reference.rst

    """

    def __init__(self, joints, ext_axes, speed, zone, feedback_level=FeedbackLevel.NONE):
        """Create a new instance of the instruction.

        Parameters
        ----------
        joints : :class:`compas_rrc.RobotJoints` or :obj:`list` of :obj:`float`
            Robot joint positions.
        ext_axes : :class:`compas_rrc.ExternalAxes` or :obj:`list` of :obj:`float`
            External axes positions.
        speed : :obj:`float`
            Integer specifying TCP translational speed in mm/s. Min=``0.01``.
        zone : :class:`Zone`
            Zone data. Predefined in the robot controller,
            only Zone :attr:`Zone.FINE` will do a stop point all others are fly by points
        feedback_level : :obj:`int`
            Defines the feedback level requested from the robot. Defaults to :attr:`FeedbackLevel.NONE`.
            Use  :attr:`FeedbackLevel.DONE` and :attr:`Zone.FINE` together to make sure
            the motion planner has executed the instruction fully.
        """
        if speed <= 0:
            raise ValueError('Speed must be higher than zero. Current value={}'.format(speed))

        self.instruction = INSTRUCTION_PREFIX + 'MoveToJoints'
        self.feedback_level = feedback_level
        self.exec_level = ExecutionLevel.ROBOT

        joints = joints or []
        if len(joints) > 6:
            raise ValueError('Only up to 6 joints are supported')
        joints_pad = [0.0] * (6 - len(joints))

        ext_axes = ext_axes or []
        if len(ext_axes) > 6:
            raise ValueError('Only up to 6 external axes are supported')

        ext_axes_pad = [0.0] * (6 - len(ext_axes))
        self.string_values = []
        self.float_values = list(joints) + joints_pad + list(ext_axes) + ext_axes_pad + [speed, zone]


class MoveGeneric(ROSmsg):
    def __init__(self, frame, ext_axes, speed, zone, feedback_level=FeedbackLevel.NONE):
        if speed <= 0:
            raise ValueError('Speed must be higher than zero. Current value={}'.format(speed))

        self.feedback_level = feedback_level
        self.exec_level = ExecutionLevel.ROBOT

        pos = list(frame.point)
        rot = list(frame.quaternion)

        ext_axes = ext_axes or []
        if len(ext_axes) > 6:
            raise ValueError('Only up to 6 external axes are supported')
        ext_axes_pad = [0.0] * (6 - len(ext_axes))

        self.string_values = []
        self.float_values = pos + rot + list(ext_axes) + ext_axes_pad + [speed, zone]


class MoveToFrame(MoveGeneric):
    """Move to frame is a call that moves the robot in cartesian space.
    The external axes (if any) will stay in the same position.

    Examples
    --------
    .. code-block:: python

        # Get frame
        frame = abb.send_and_wait(rrc.GetFrame())

        # Print received values
        print(frame)

        # Change any frame value and move robot to new position
        frame.point[0] += 50
        speed = 100 # Unit [mm/s]
        done = abb.send_and_wait(rrc.MoveToFrame(frame, speed, rrc.Zone.FINE, rrc.Motion.LINEAR))

    RAPID Instruction: ``MoveJ`` or ``MoveL``

    .. include:: ../abb-reference.rst

    """

    def __init__(self, frame, speed, zone, motion_type=Motion.JOINT, feedback_level=FeedbackLevel.NONE):
        """Create a new instance of the instruction.

        Parameters
        ----------
        frame : :class:`compas.geometry.Frame`
            Target frame.
        speed : :obj:`float`
            Integer specifying TCP translational speed in mm/s. Min=``0.01``.
        zone : :class:`Zone`
            Zone data. Predefined in the robot controller,
            only Zone :attr:`Zone.FINE` will do a stop point all others are fly by points
        motion_type : :class:`Motion`
            Motion type. Defaults to :attr:`Motion.JOINT`.
        feedback_level : :obj:`int`
            Defines the feedback level requested from the robot. Defaults to :attr:`FeedbackLevel.NONE`.
            Use  :attr:`FeedbackLevel.DONE` and :attr:`Zone.FINE` together to make sure
            the motion planner has executed the instruction fully.
        """
        super(MoveToFrame, self).__init__(frame, [], speed, zone, feedback_level)
        instruction = 'MoveTo'
        self.instruction = INSTRUCTION_PREFIX + instruction
        self.string_values = ['FrameJ'] if motion_type == Motion.JOINT else ['FrameL']


class MoveToRobtarget(MoveGeneric):
    """Move to robtarget is a call that moves the robot in cartesian space with explicit external axes values.

    Examples
    --------
    .. code-block:: python

        # Get frame and external axes
        frame, external_axes = abb.send_and_wait(rrc.GetRobtarget())

        # Print received values
        print(frame, external_axes)

        # Change any value and move to new position
        frame.point[0] += 50
        speed = 100 # Unit [mm/s]
        done = abb.send_and_wait(rrc.MoveToRobtarget(frame, external_axes, speed, rrc.Zone.FINE))

    RAPID Instruction: ``MoveJ`` or ``MoveL``

    .. include:: ../abb-reference.rst

    """

    def __init__(self, frame, ext_axes, speed, zone, motion_type=Motion.JOINT, feedback_level=FeedbackLevel.NONE):
        """Create a new instance of the instruction.

        Parameters
        ----------
        frame : :class:`compas.geometry.Frame`
            Target frame.
        ext_axes : :class:`compas_rrc.ExternalAxes` or :obj:`list` of :obj:`float`
            External axes positions.
        speed : :obj:`float`
            Integer specifying TCP translational speed in mm/s. Min=``0.01``.
        zone : :class:`Zone`
            Zone data. Predefined in the robot controller,
            only Zone :attr:`Zone.FINE` will do a stop point all others are fly by points
        motion_type : :class:`Motion`
            Motion type. Defaults to :attr:`Motion.JOINT`.
        feedback_level : :obj:`int`
            Defines the feedback level requested from the robot. Defaults to :attr:`FeedbackLevel.NONE`.
            Use  :attr:`FeedbackLevel.DONE` and :attr:`Zone.FINE` together to make sure
            the motion planner has executed the instruction fully.
        """
        super(MoveToRobtarget, self).__init__(frame, ext_axes, speed, zone, feedback_level)
        instruction = 'MoveTo'
        self.instruction = INSTRUCTION_PREFIX + instruction
        self.string_values = ['J'] if motion_type == Motion.JOINT else ['L']



from compas_fab.backends.ros.messages import ROSmsg

from compas_rrc.common import ExecutionLevel
from compas_rrc.common import FeedbackLevel

INSTRUCTION_PREFIX = 'r_RRC_'

__all__ = [
    'SetDigital',
    'SetAnalog',
    'SetGroup',
    'PulseDigital',
    'ReadAnalog',
    'ReadDigital',
    'ReadGroup',
]


class SetDigital(ROSmsg):
    """Set digital is a call that sets the value of an digital output signal (:obj:`int`, 0 or 1).

    Examples
    --------
    .. code-block:: python

        # Set digital output to low or high
        low = 0
        high = 1
        done = abb.send_and_wait(rrc.SetDigital('do_1',low))

    RAPID Instruction: ``SetDO``

    .. include:: ../abb-reference.rst

    """

    def __init__(self, io_name, value, feedback_level=FeedbackLevel.NONE):
        """Create a new instance of the instruction.

        Parameters
        ----------
        io_name : :obj:`str`
            Name of the output signal.
        value : :obj:`int`
            Value of the output signal.
        feedback_level : :obj:`int`
            Defines the feedback level requested from the robot. Defaults to :attr:`FeedbackLevel.NONE`.
        """
        if value not in (0, 1):
            raise ValueError("Value can only be 0 or 1")
        self.instruction = INSTRUCTION_PREFIX + 'SetDigital'
        self.feedback_level = feedback_level
        self.exec_level = ExecutionLevel.ROBOT
        self.string_values = [io_name]
        self.float_values = [value]


class SetAnalog(ROSmsg):
    """Set analog is a call that sets the value of an analog output signal (:obj:`float`).
    Minimum and maximum values are given by the signal configuration in the robot.

    Examples
    --------
    .. code-block:: python

        # Set analog output
        value = -3.33
        done = abb.send_and_wait(rrc.SetAnalog('ao_1', value))

    RAPID Instruction: ``SetAO``

    .. include:: ../abb-reference.rst

    """

    def __init__(self, io_name, value, feedback_level=FeedbackLevel.NONE):
        """Create a new instance of the instruction.

        Parameters
        ----------
        io_name : :obj:`str`
            Name of the output signal.
        value : :obj:`float`
            Value of the output signal.
        feedback_level : :obj:`int`
            Defines the feedback level requested from the robot. Defaults to :attr:`FeedbackLevel.NONE`.
        """
        self.instruction = INSTRUCTION_PREFIX + 'SetAnalog'
        self.feedback_level = feedback_level
        self.exec_level = ExecutionLevel.ROBOT
        self.string_values = [io_name]
        self.float_values = [value]


class SetGroup(ROSmsg):
    """Set group is a call that sets the value of an digital group output signal (:obj:`int`).
    Minimum and maximum values are given by the signal configuration in the robot.

    Examples
    --------
    .. code-block:: python

        # Set group output
        value = 33
        done = abb.send_and_wait(rrc.SetGroup('go_1', value))

    RAPID Instruction: ``SetGO``

    .. include:: ../abb-reference.rst

    """

    def __init__(self, io_name, value, feedback_level=FeedbackLevel.NONE):
        """Create a new instance of the instruction.

        Parameters
        ----------
        io_name : :obj:`str`
            Name of the output signal.
        value : :obj:`int`
            Value of the output signal.
        feedback_level : :obj:`int`
            Defines the feedback level requested from the robot. Defaults to :attr:`FeedbackLevel.NONE`.
        """
        self.instruction = INSTRUCTION_PREFIX + 'SetGroup'
        self.feedback_level = feedback_level
        self.exec_level = ExecutionLevel.ROBOT
        self.string_values = [io_name]
        self.float_values = [value]


class PulseDigital(ROSmsg):
    """Pulse digital is a call that sets the value to HIGH of an digital output signal for a certain time.

    Examples
    --------
    .. code-block:: python

        # Pulse digital output
        pulse_time = 2.5 # Unit [s]
        done = abb.send_and_wait(rrc.PulseDigital('do_1', pulse_time))

    RAPID Instruction: ``PulseDO``

    .. include:: ../abb-reference.rst

    """

    def __init__(self, io_name, pulse_time, feedback_level=FeedbackLevel.NONE):
        """Create a new instance of the instruction.

        Parameters
        ----------
        io_name : :obj:`str`
            Name of the output signal.
        pulse_time : :obj:`float`
            Duration of the pulse expressed in seconds.
        feedback_level : :obj:`int`
            Defines the feedback level requested from the robot. Defaults to :attr:`FeedbackLevel.NONE`.
        """
        self.instruction = INSTRUCTION_PREFIX + 'PulseDigital'
        self.feedback_level = feedback_level
        self.exec_level = ExecutionLevel.ROBOT
        self.string_values = [io_name]
        self.float_values = [pulse_time]


class ReadAnalog(ROSmsg):
    """Read analog is a call that requests the value of an analog input signal.

    Examples
    --------
    .. code-block:: python

        # Read analog
        analog_input_1 = abb.send_and_wait(rrc.ReadAnalog('ai_1'))

    RAPID Instruction: ``AInput``

    .. include:: ../abb-reference.rst

    """

    def __init__(self, io_name):
        """Create a new instance of the instruction.

        Parameters
        ----------
        io_name : :obj:`str`
            Name of the input signal.
        """
        self.instruction = INSTRUCTION_PREFIX + 'ReadAnalog'
        self.feedback_level = FeedbackLevel.DONE
        self.exec_level = ExecutionLevel.ROBOT
        self.string_values = [io_name]

    def parse_feedback(self, result):
        """Parses the result as a :obj:`float`.

        Return
        ------
        :obj:`float`
            Current value of the input signal.
        """
        # read input value
        result = result['float_values'][0]
        return result


class ReadDigital(ROSmsg):
    """Read digital is a call that requests the value of a digital input signal.

    Examples
    --------
    .. code-block:: python

        # Read digital
        digital_input_1 = abb.send_and_wait(rrc.ReadDigital('di_1'))

    RAPID Instruction: ``DInput``

    .. include:: ../abb-reference.rst

    """

    def __init__(self, io_name):
        """Create a new instance of the instruction.

        Parameters
        ----------
        io_name : :obj:`str`
            Name of the input signal.
        """
        self.instruction = INSTRUCTION_PREFIX + 'ReadDigital'
        self.feedback_level = FeedbackLevel.DONE
        self.exec_level = ExecutionLevel.ROBOT
        self.string_values = [io_name]

    def parse_feedback(self, result):
        """Parses the result as a :obj:`int`.

        Return
        ------
        :obj:`int`
            Current value of the input signal.
        """

        # read input value
        result = int(result['float_values'][0])
        return result


class ReadGroup(ROSmsg):
    """Read group is a call that requests the value of a digital group input signal.

    Examples
    --------
    .. code-block:: python

        # Read group
        group_input_1 = abb.send_and_wait(rrc.ReadGroup('gi_1'))

    RAPID Instruction: ``GInput``

    .. include:: ../abb-reference.rst

    """

    def __init__(self, io_name):
        """Create a new instance of the instruction.

        Parameters
        ----------
        io_name : :obj:`str`
            Name of the input signal.
        """
        self.instruction = INSTRUCTION_PREFIX + 'ReadGroup'
        self.feedback_level = FeedbackLevel.DONE
        self.exec_level = ExecutionLevel.ROBOT
        self.string_values = [io_name]

    def parse_feedback(self, result):
        """Parses the result as a :obj:`int`.

        Return
        ------
        :obj:`int`
            Current value of the input signal.
        """

        # read input value
        result = int(result['float_values'][0])
        return result



import pytest
from compas.geometry import Frame

import compas_rrc as rrc


def test_move_to_joints():
    joints = [30, 90, 0, 0, 0]
    extaxe = [0, 0, 100]
    inst = rrc.MoveToJoints(joints, extaxe, 100, rrc.Zone.FINE)

    assert inst.float_values == [30, 90, 0, 0, 0, 0, 0, 0, 100, 0, 0, 0, 100, -1]
    assert not inst.string_values
    assert inst.exec_level == rrc.ExecutionLevel.ROBOT
    assert inst.feedback_level == rrc.FeedbackLevel.NONE

    joints = rrc.RobotJoints(30, 90, 0, 0, 0)
    extaxe = rrc.ExternalAxes(0, 0, 100)
    inst = rrc.MoveToJoints(joints, extaxe, 100, rrc.Zone.FINE)

    assert inst.float_values == [30, 90, 0, 0, 0, 0, 0, 0, 100, 0, 0, 0, 100, -1]


def test_move_to_joints_validation():
    # Only up to 6 joints are supported
    with pytest.raises(ValueError):
        rrc.MoveToJoints([30, 90, 0, 0, 0, 0, 0, 0], [0, 0, 100], 100, rrc.Zone.FINE)

    # Only up to 6 external axes are supported
    with pytest.raises(ValueError):
        rrc.MoveToJoints([30, 90], [0, 0, 100, 0, 0, 0, 0], 100, rrc.Zone.FINE)


def test_move_to_frame():
    inst = rrc.MoveToFrame(Frame.worldXY(), 100, rrc.Zone.FINE, rrc.Motion.JOINT)

    assert inst.float_values == [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0, 0, 0, 0, 0, 0, 100, -1]
    assert inst.string_values == ['FrameJ']

    inst = rrc.MoveToFrame(Frame.worldXY(), 100, rrc.Zone.FINE, rrc.Motion.LINEAR)
    assert inst.string_values == ['FrameL']


def test_move_to_robtarget():
    inst = rrc.MoveToRobtarget(Frame.worldXY(), [50, 20], 100, rrc.Zone.FINE, rrc.Motion.JOINT)

    assert inst.float_values == [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 50, 20, 0, 0, 0, 0, 100, -1]
    assert inst.string_values == ['J']


def test_move_to_robtarget_validation():
    # Only up to 6 external axes are supported
    with pytest.raises(ValueError):
        rrc.MoveToRobtarget(Frame.worldXY(), [50, 20, 0, 0, 0, 0, 0, 0], 100, rrc.Zone.FINE, rrc.Motion.JOINT)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os
import sphinx_compas_theme

from sphinx.ext.napoleon.docstring import NumpyDocstring

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.coverage',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.extlinks',
    'sphinx.ext.ifconfig',
    'sphinx.ext.napoleon',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
]
if os.getenv('SPELLCHECK'):
    extensions += 'sphinxcontrib.spelling',
    spelling_show_suggestions = True
    spelling_lang = 'en_US'

source_suffix = '.rst'
master_doc = 'index'
project = 'COMPAS RRC'
year = '2019'
author = 'ETH Zurich'
copyright = '{0}, {1}'.format(year, author)
version = release = '1.1.0'

pygments_style = 'sphinx'
show_authors = True
add_module_names = True
templates_path = ['_templates', ]
extlinks = {
    'issue': ('https://github.com/compas-rrc/compas_rrc/issues/%s', '#'),
    'pr': ('https://github.com/compas-rrc/compas_rrc/pull/%s', 'PR #'),
}

# intersphinx options
intersphinx_mapping = {'python': ('https://docs.python.org/', None),
                       'compas': ('https://compas.dev/compas/latest/', None),
                       'compas_fab': ('https://gramaziokohler.github.io/compas_fab/latest/', None),
                       'roslibpy': ('https://roslibpy.readthedocs.io/en/latest/', None),
                       }

# autodoc options
autodoc_default_options = {
    'member-order': 'bysource',
    'exclude-members': '__weakref__',
    'undoc-members': True,
    'private-members': True,
    'show-inheritance': True,
}

autodoc_member_order = 'alphabetical'

# autosummary options
autosummary_generate = True

# collect doc versions
package_docs_root = 'https://compasrrc.github.io/compas_rrc/'

with open(os.path.join(os.path.dirname(__file__), 'doc_versions.txt'), 'r') as f:
    version_names = [version.strip() for version in f.readlines()]
    package_docs_versions = [(version, '{}{}'.format(package_docs_root, version))
                             for version in version_names if version]

# on_rtd is whether we are on readthedocs.org
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
html_theme = 'compaspkg'
html_theme_path = sphinx_compas_theme.get_html_theme_path()
html_theme_options = {
    "package_name": 'compas_rrc',
    "package_title": project,
    "package_version": release,
    "package_repo": 'https://github.com/compas-rrc/compas_rrc',
    "package_docs": package_docs_root,
    "package_old_versions": package_docs_versions
}

html_split_index = False
html_short_title = '%s-%s' % (project, version)
html_context = {}
html_static_path = ['_static']
html_last_updated_fmt = '%b %d, %Y'
html_copy_source = False
html_show_sourcelink = False
html_add_permalinks = ''
html_permalinks = False
html_experimental_html5_writer = True
html_compact_lists = True

# napoleon options
napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = True
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = False
napoleon_use_rtype = False

# Parse Attributes and Class Attributes on Class docs same as parameters.
# first, we define new methods for any new sections and add them to the class


def parse_keys_section(self, section):
    return self._format_fields('Keys', self._consume_fields())


NumpyDocstring._parse_keys_section = parse_keys_section


def parse_attributes_section(self, section):
    return self._format_fields('Attributes', self._consume_fields())


NumpyDocstring._parse_attributes_section = parse_attributes_section


def parse_class_attributes_section(self, section):
    return self._format_fields('Class Attributes', self._consume_fields())


NumpyDocstring._parse_class_attributes_section = parse_class_attributes_section


# we now patch the parse method to guarantee that the the above methods are
# assigned to the _section dict
def patched_parse(self):
    self._sections['keys'] = self._parse_keys_section
    self._sections['class attributes'] = self._parse_class_attributes_section
    self._unpatched_parse()


NumpyDocstring._unpatched_parse = NumpyDocstring._parse
NumpyDocstring._parse = patched_parse



from compas_fab.backends.ros.messages import ROSmsg

from compas_rrc.common import ExecutionLevel
from compas_rrc.common import FeedbackLevel

__all__ = ['CustomInstruction']


class CustomInstruction(ROSmsg):
    """Custom instruction is a call that invokes a custom RAPID instruction. The name has to match a ``RAPID`` procedure.

    Examples
    --------
    .. code-block:: python

        # Custom instruction
        string_values = ['Custom Text']
        float_values = [42]
        done = abb.send_and_wait(rrc.CustomInstruction('r_RRC_CustomInstruction', string_values, float_values))

    RAPID Instruction: ``All usable``

    .. include:: ../abb-reference.rst

    """

    def __init__(self, name, string_values=[], float_values=[], feedback_level=FeedbackLevel.NONE, execution_level=ExecutionLevel.ROBOT):
        """Create a new instance of the instruction.

        Parameters
        ----------
        name : :obj:`str`
            Name of the procedure to invoke on the robot code. Maximum of 80 characters.
        string_values : :obj:`list` of :obj:`str`
            List of up to 8 strings values, each of them with a maximum of 80 characters.
        float_values : :obj:`list` of :obj:`float`
            List of up to 36 float values.
        feedback_level : :obj:`int`
            Defines the feedback level requested from the robot. Defaults to :attr:`FeedbackLevel.NONE`.
        execution_level : :obj:`int`
            Defines the execution level of the instruction. Defaults to :attr:`ExecutionLevel.ROBOT`.
        """
        self.instruction = name
        self.feedback_level = feedback_level
        self.exec_level = execution_level
        self.string_values = string_values
        self.float_values = float_values



__title__ = 'compas_rrc'
__description__ = 'COMPAS RRC: Library for ABB Robots'
__url__ = 'https://github.com/compas-rrc/compas_rrc'
__version__ = '1.1.0'
__author__ = 'ETH Zurich'
__author_email__ = 'rrc@arch.ethz.ch'
__license__ = 'MIT license'
__copyright__ = 'Copyright 2019 ETH Zurich'

__all__ = ['__author__', '__author_email__', '__copyright__', '__description__', '__license__', '__title__', '__url__', '__version__']



from __future__ import print_function

import os

import pytest

HERE = os.path.dirname(__file__)

if __name__ == '__main__':
    # Fake Rhino modules
    pytest.load_fake_module('Rhino')
    pytest.load_fake_module('Rhino.Geometry', fake_types=['RTree', 'Sphere', 'Point3d'])

    pytest.run(HERE)



from compas_fab.backends.ros.messages import ROSmsg

from compas_rrc.common import ExecutionLevel
from compas_rrc.common import FeedbackLevel

INSTRUCTION_PREFIX = 'r_RRC_'

__all__ = [
    'ReadWatch',
    'StartWatch',
    'StopWatch',
]


class ReadWatch(ROSmsg):
    """Read Watch is a call that requests the value of the watch in the robot code.

    Examples
    --------
    .. code-block:: python

        # Read watch
        watch_time = abb.send_and_wait(rrc.ReadWatch())  # Unit [s]

    RAPID Instruction: ``ClkRead``

    .. include:: ../abb-reference.rst

    """

    def __init__(self):
        """Create a new instance of the instruction."""
        self.instruction = INSTRUCTION_PREFIX + 'ReadWatch'
        self.feedback_level = FeedbackLevel.DONE
        self.exec_level = ExecutionLevel.ROBOT
        self.string_values = []
        self.float_values = []

    def parse_feedback(self, result):
        """Parses the result as a :obj:`float` (seconds).

        Return
        ------
        :obj:`float`
            Current value of the watch in seconds.
        """
        # read input value
        result = round(result['float_values'][0], 3)
        return result


class StartWatch(ROSmsg):
    """Start Watch is a call that starts the watch in the robot code.

    Examples
    --------
    .. code-block:: python

        # Start watch
        done = abb.send_and_wait(rrc.StartWatch())

    RAPID Instruction: ``ClkStart``

    .. include:: ../abb-reference.rst

    """

    def __init__(self, feedback_level=FeedbackLevel.NONE):
        """Create a new instance of the instruction.

        Parameters
        ----------
        feedback_level : :obj:`int`
            Defines the feedback level requested from the robot. Defaults to :attr:`FeedbackLevel.NONE`.
        """
        self.instruction = INSTRUCTION_PREFIX + 'StartWatch'
        self.feedback_level = feedback_level
        self.exec_level = ExecutionLevel.ROBOT
        self.string_values = []
        self.float_values = []


class StopWatch(ROSmsg):
    """Stop Watch is a call that stops the watch in the robot code.

    Examples
    --------
    .. code-block:: python

        # Stop watch
        done = abb.send_and_wait(rrc.StopWatch())

    RAPID Instruction: ``ClkStop``

    .. include:: ../abb-reference.rst

    """

    def __init__(self, feedback_level=FeedbackLevel.NONE):
        """Create a new instance of the instruction.

        Parameters
        ----------
        feedback_level : :obj:`int`
            Defines the feedback level requested from the robot. Defaults to :attr:`FeedbackLevel.NONE`.
        """
        self.instruction = INSTRUCTION_PREFIX + 'StopWatch'
        self.feedback_level = feedback_level
        self.exec_level = ExecutionLevel.ROBOT
        self.string_values = []
        self.float_values = []



from compas_fab.backends.ros.messages import ROSmsg

from compas_rrc.common import ExecutionLevel
from compas_rrc.common import FeedbackLevel

INSTRUCTION_PREFIX = 'r_RRC_'

__all__ = [
    'PrintText'
]


class PrintText(ROSmsg):
    """Print text is a call that prints a single line of text on the robot panel.

    Examples
    --------
    .. code-block:: python

        # Print text
        done = abb.send_and_wait(rrc.PrintText('Welcome to COMPAS_RRC'))

    RAPID Instruction: ``TPWrite``

    .. include:: ../abb-reference.rst

    """

    def __init__(self, text, feedback_level=FeedbackLevel.NONE):
        """Create a new instance of the instruction.

        Parameters
        ----------
        text : :obj:`str`
            Single line of text to print on the robot panel with a maximum of 80 characters.
        feedback_level : :obj:`int`
            Defines the feedback level requested from the robot. Defaults to :attr:`FeedbackLevel.NONE`.
        """
        if len(text) > 80:
            raise ValueError("text can only be up to 80 chars")
        self.instruction = INSTRUCTION_PREFIX + 'PrintText'
        self.feedback_level = feedback_level
        self.exec_level = ExecutionLevel.ROBOT
        self.string_values = [text]
        self.float_values = []



import compas_rrc as rrc


def test_robot_joints():
    j = rrc.RobotJoints()
    assert list(j) == []

    j = rrc.RobotJoints(30)
    assert list(j) == [30]

    j = rrc.RobotJoints(30, 10, 0)
    assert list(j) == [30, 10, 0]

    j = rrc.RobotJoints([30, 10, 0])
    assert list(j) == [30, 10, 0]

    j = rrc.RobotJoints(iter([30, 10, 0]))
    assert list(j) == [30, 10, 0]


def test_external_axes():
    ea = rrc.ExternalAxes()
    assert list(ea) == []

    ea = rrc.ExternalAxes(30)
    assert list(ea) == [30]

    ea = rrc.ExternalAxes(30, 10, 0)
    assert list(ea) == [30, 10, 0]

    ea = rrc.ExternalAxes([30, 10, 0])
    assert list(ea) == [30, 10, 0]

    ea = rrc.ExternalAxes(iter([30, 10, 0]))
    assert list(ea) == [30, 10, 0]



import itertools
import threading

__all__ = ['CLIENT_PROTOCOL_VERSION',
           'FeedbackLevel',
           'ExecutionLevel',
           'InstructionException',
           'TimeoutException',
           'FutureResult',
           'ExternalAxes',
           'RobotJoints']


CLIENT_PROTOCOL_VERSION = 2


class FeedbackLevel(object):
    """Represents default valid feedback levels.

    .. autoattribute:: NONE
    .. autoattribute:: DONE
    """
    NONE = 0
    """Indicates no feedback is requested from the robot."""

    DONE = 1
    """Indicates completion feedback is requested from the robot. Completion feedback means
    the robot has executed the procedure. See :meth:`AbbClient.send_and_wait` for more details.
    """


class ExecutionLevel(object):
    """Defines the execution level of an instruction.

    .. autoattribute:: ROBOT
    .. autoattribute:: CONTROLLER
    """
    ROBOT = 0
    """Execute instruction on the robot task."""

    CONTROLLER = 10
    """Execute instruction on the ``controller`` task (only usable with custom instructions)."""


class InstructionException(Exception):
    """Exception caused during/after the execution of an instruction."""

    def __init__(self, message, result):
        super(InstructionException, self).__init__('{}, RRC Reply={}'.format(message, result))
        self.result = result


class TimeoutException(Exception):
    """Timeout exception caused during execution of an instruction."""
    pass


class FutureResult(object):
    """Represents a future result value.

    Futures are the result of asynchronous operations
    but allow to explicitely control when to block and wait
    for its completion."""

    def __init__(self):
        self.done = False
        self.value = None
        self.event = threading.Event()

    def result(self, timeout=None):
        """Return the feedback value returned by the instruction.

        If the instruction has not yet returned feedback, it will wait
        up to ``timeout`` seconds. If the ``timeout`` expires, the method
        will raise an exception.
        """
        if not self.done:
            if not self.event.wait(timeout):
                raise TimeoutException('Timeout: future result not available')

        if isinstance(self.value, Exception):
            raise self.value

        return self.value

    def _set_result(self, value):
        self.value = value
        self.done = True
        self.event.set()


class ExternalAxes(object):
    """Represents a configuration for external axes."""

    def __init__(self, *values):
        """Initialize a new object with the specified values for external axes.

        Parameters
        ----------
        values : :obj:`list`
            List of floats indicating the external axis positions.
        """
        try:
            self.values = list(itertools.chain.from_iterable(values))
        except TypeError:
            self.values = list(values)

    # Properties
    @property
    def eax_a(self):
        """Value of the first external axis."""
        return self[0]

    @eax_a.setter
    def eax_a(self, value):
        self[0] = value

    @property
    def eax_b(self):
        """Value of the second external axis."""
        return self[1]

    @eax_b.setter
    def eax_b(self, value):
        self[1] = value

    @property
    def eax_c(self):
        """Value of the third external axis."""
        return self[2]

    @eax_c.setter
    def eax_c(self, value):
        self[2] = value

    @property
    def eax_d(self):
        """Value of the fourth external axis."""
        return self[3]

    @eax_d.setter
    def eax_d(self, value):
        self[3] = value

    @property
    def eax_e(self):
        """Value of the fifth external axis."""
        return self[4]

    @eax_e.setter
    def eax_e(self, value):
        self[4] = value

    @property
    def eax_f(self):
        """Value of the sexth external axis."""
        return self[5]

    @eax_f.setter
    def eax_f(self, value):
        self[5] = value

    # List accessors
    def __repr__(self):
        return 'ExternalAxes({})'.format([round(i, 2) for i in self.values])

    def __len__(self):
        return len(self.values)

    def __getitem__(self, item):
        if item >= len(self.values):
            return None

        return self.values[item]

    def __setitem__(self, item, value):
        self.values[item] = value

    def __iter__(self):
        return iter(self.values)


class RobotJoints(object):
    """Represents a configuration for robot joints"""

    def __init__(self, *values):
        try:
            self.values = list(itertools.chain.from_iterable(values))
        except TypeError:
            self.values = list(values)

    # Properties
    @property
    def rax_1(self):
        return self[0]

    @rax_1.setter
    def rax_1(self, value):
        self[0] = value

    @property
    def rax_2(self):
        return self[1]

    @rax_2.setter
    def rax_2(self, value):
        self[1] = value

    @property
    def rax_3(self):
        return self[2]

    @rax_3.setter
    def rax_3(self, value):
        self[2] = value

    @property
    def rax_4(self):
        return self[3]

    @rax_4.setter
    def rax_4(self, value):
        self[3] = value

    @property
    def rax_5(self):
        return self[4]

    @rax_5.setter
    def rax_5(self, value):
        self[4] = value

    @property
    def rax_6(self):
        return self[5]

    @rax_6.setter
    def rax_6(self, value):
        self[5] = value

    # List accessors
    def __repr__(self):
        return 'RobotJoints({})'.format([round(i, 2) for i in self.values])

    def __len__(self):
        return len(self.values)

    def __getitem__(self, item):
        if item >= len(self.values):
            return None

        return self.values[item]

    def __setitem__(self, item, value):
        self.values[item] = value

    def __iter__(self):
        return iter(self.values)



import threading
import time

import roslibpy
from compas_fab.backends import RosClient

from .common import CLIENT_PROTOCOL_VERSION
from .common import FutureResult
from .common import InstructionException

__all__ = ['RosClient', 'AbbClient']


FEEDBACK_ERROR_PREFIX = 'Done FError '


def _get_key(message):
    return 'msg:{}'.format(message.sequence_id)


def _get_response_key(message):
    return 'msg:{}'.format(message['feedback_id'])


class SequenceCounter(object):
    """An atomic, thread-safe sequence increament counter."""

    def __init__(self, start=0):
        """Initialize a new counter to given initial value."""
        self._lock = threading.Lock()
        self._value = start

    def increment(self, num=1):
        """Atomically increment the counter by ``num`` and
        return the new value.
        """
        with self._lock:
            self._value += num
            return self._value

    @property
    def value(self):
        """Current sequence counter."""
        with self._lock:
            return self._value


def default_feedback_parser(result):
    feedback_value = result['feedback']

    if feedback_value.startswith(FEEDBACK_ERROR_PREFIX):
        return InstructionException(feedback_value, result)

    return feedback_value


class AbbClient(object):
    """Client used to communicate with ABB robots via ROS.

    This client handles all communication over ROS topics, and implements
    blocking behaviors as an application-level construct.

    Examples
    --------

    Connection example to a single robot::

        # Create Ros Client
        ros = rrc.RosClient()
        ros.run()

        # Create ABB Client
        abb = rrc.AbbClient(ros, '/rob1')
        print('Connected.')

        # Close client
        ros.close()
        ros.terminate()

    Advance connection example to multiple robots::

        # Create Ros Client
        ros = rrc.RosClient()
        ros.run()

        # Create ABB Clients
        abb_rob1 = rrc.AbbClient(ros, '/rob1')
        abb_rob2 = rrc.AbbClient(ros, '/rob2')

        # Clients are connected
        print('Connected.')

        # Print Text
        abb_rob1.send(rrc.PrintText('Hello Robot 1'))
        abb_rob2.send(rrc.PrintText('Hello Robot 2'))

        # Close client
        ros.close()
        ros.terminate()

    """

    def __init__(self, ros, namespace='/rob1'):
        """Initialize a new robot client instance.

        Parameters
        ----------
        ros : :class:`RosClient`
            Instance of a ROS connection.
        namespace : :obj:`str`
            Namespace to allow multiple robots to be controlled through the same ROS instance.
            Optional. If not specified, it will use namespace ``/rob1``.
        """
        self.ros = ros
        self.counter = SequenceCounter()
        if not namespace.endswith('/'):
            namespace += '/'
        self._version_checked = False
        self._server_protocol_check = dict(event=threading.Event(),
                                           param=roslibpy.Param(ros, namespace + 'protocol_version'),
                                           version=None)
        self.ros.on_ready(self.version_check)
        self.topic = roslibpy.Topic(ros, namespace + 'robot_command', 'compas_rrc_driver/RobotMessage', queue_size=None)
        self.feedback = roslibpy.Topic(ros, namespace + 'robot_response', 'compas_rrc_driver/RobotMessage', queue_size=0)
        self.feedback.subscribe(self.feedback_callback)
        self.topic.advertise()
        self.futures = {}

        self.ros.on('closing', self._disconnect_topics)

    def version_check(self):
        """Check if the protocol version on the server matches the protocol version on the client."""
        self._server_protocol_check['version'] = self._server_protocol_check['param'].get()
        self._server_protocol_check['event'].set()

    def ensure_protocol_version(self):
        """Ensure protocol version on the server matches the protocol version on the client."""
        if self._version_checked:
            return

        if not self._server_protocol_check['version']:
            if not self._server_protocol_check['event'].wait(10):
                raise Exception('Could not yet retrieve server protocol version')

        if self._server_protocol_check['version'] != CLIENT_PROTOCOL_VERSION:
            raise Exception('Protocol version mismatch. Server={}, Client={}'.format(self._server_protocol_check['version'], CLIENT_PROTOCOL_VERSION))

        self._version_checked = True

    def _disconnect_topics(self):
        self.topic.unadvertise()
        self.feedback.unsubscribe()
        time.sleep(0.5)

    def send(self, instruction):
        """Sends an instruction to the robot without waiting.

        Instructions can indicate that feedback is required or not. If
        the instruction sent does not require feedback, this method
        returns ``None``. However, if the instruction needs
        feedback (i.e. ``feedback_level`` is greater than zero), the method
        returns a future result object that can be used to wait for completion.

        Waiting for a future can be done immediately after calling this, or
        deferred to a later point.

        Parameters
        ----------
        instruction : :class:`compas_fab.backends.ros.messages.ROSmsg`
            ROS Message representing the instruction to send.

        Returns
        -------
        :class:`FutureResult`:
            Represent the future value of the feedback request. This method
            will return immediately, and this object can be used to wait or
            react to the feedback whenever it becomes available.

        Examples
        --------

        Streaming commands without blocking or waiting for feedback::

            # Print path
            abb.send(rrc.MoveToFrame(Frame.worldXY(), 150, rrc.Zone.FINE, rrc.Motion.LINEAR))

        Send commands and defer waiting to a future point in time::

            # Stop watch
            done = abb.send_and_wait(rrc.StopWatch())

            # Read watch
            future = abb.send(rrc.ReadWatch())

            # Move robot to end position
            abb.send(rrc.MoveToJoints(robot_joints_end_position, external_axis_dummy, 1000, rrc.Zone.FINE))

            # Read and print printing time
            watch_time = future.result(timeout=3.0)
            print('Print Time [s] = ', watch_time)

        """
        self.ensure_protocol_version()
        instruction.sequence_id = self.counter.increment()

        key = _get_key(instruction)
        result = None

        if instruction.feedback_level > 0:
            result = FutureResult()
            parser = instruction.parse_feedback if hasattr(instruction, 'parse_feedback') else None
            self.futures[key] = dict(result=result, parser=parser)

        self.topic.publish(roslibpy.Message(instruction.msg))

        return result

    def send_and_wait(self, instruction, timeout=None):
        """Send instruction and wait for feedback.

        This is a blocking call, it will only return once the robot
        sends the requested feedback. If ``feedback_level``
        of the ``instruction`` parameter is ``0``, it will be automatically
        set to ``1``.

        Parameters
        ----------
        instruction : :class:`compas_fab.backends.ros.messages.ROSmsg`
            ROS Message representing the instruction to send.
        timeout : :obj:`int`
            Timeout in seconds to wait before raising an exception. Optional.

        Returns
        -------
        object
            Returns the feedback value that resulted from the execution of the instruction.

        Examples
        --------

        Send an instruction and wait for feedback from the robot. In the following example,
        the code will not continue until the robot has started to execute this instruction.
        On move instructions, a ``Zone.FINE`` can be used to make sure the motion planner has
        executed the instruction fully::

            # Move robot to start position
            done = abb.send_and_wait(rrc.MoveToJoints(robot_joints_start_position, external_axis_dummy, 1000, rrc.Zone.FINE))

        """
        if instruction.feedback_level == 0:
            instruction.feedback_level = 1

        future = self.send(instruction)
        return future.result(timeout)

    def send_and_subscribe(self, instruction, callback):
        """Send instruction and activate a service on the robot to stream feedback at a regular inverval.

        Parameters
        ----------
        instruction : :class:`compas_fab.backends.ros.messages.ROSmsg`
            ROS Message representing the instruction to send.
        callback
            Python function to be invoked every time a new value is made available.

        Notes
        -----
            This feature is currently only usable with custom instructions.

        """
        self.ensure_protocol_version()
        instruction.sequence_id = self.counter.increment()

        key = _get_key(instruction)

        parser = instruction.parse_feedback if hasattr(instruction, 'parse_feedback') else None
        self.futures[key] = dict(callback=callback, parser=parser)

        self.topic.publish(roslibpy.Message(instruction.msg))

    def feedback_callback(self, message):
        """Internal method."""
        response_key = _get_response_key(message)
        future = self.futures.get(response_key, None)

        if future:
            result = message
            if future['parser']:
                result = future['parser'](result)
            else:
                result = default_feedback_parser(result)
            if 'result' in future:
                future['result']._set_result(result)
                self.futures.pop(response_key)
            elif 'callback' in future:
                future['callback'](result)
                # TODO: Handle unsubscribes



# isort: skip_file
"""

.. currentmodule:: compas_rrc

Main concepts
=============

The API of ``COMPAS RRC`` is minimal and very easy to understand.

Communication methods
---------------------

The primary way to interact with robots is using the client classes. They allow four
different ways of communication:

* **Send**: The method :meth:`~compas_rrc.AbbClient.send` allows streaming
  commands without blocking or waiting for feedback.
* **Send & Wait**: The method :meth:`~compas_rrc.AbbClient.send_and_wait` sends
  an instruction and wait for feedback from the robot.
* **Send & Wait in the future**: Using the return value of the method
  :meth:`~compas_rrc.AbbClient.send` allows to defer the waiting to a future point in time.
* **Send & Subscribe**: The method :meth:`~compas_rrc.AbbClient.send_and_subscribe` can activate
  a streaming service on the robot that will stream feedback at a regular inverval.

.. autosummary::
    :toctree: generated/
    :nosignatures:

    RosClient
    AbbClient
    ExecutionLevel
    FeedbackLevel
    FutureResult

Robot joints and External axes
------------------------------

The following example shows how to retrieve, update and send the robot joints and external axes::

    # Get joints
    robot_joints, external_axes = abb.send_and_wait(rrc.GetJoints())

    # Print received values
    print(robot_joints, external_axes)

    # Change any value and move to new position
    robot_joints.rax_1 += 15
    done = abb.send_and_wait(rrc.MoveToJoints(robot_joints, external_axes, 100, rrc.Zone.FINE))


.. autosummary::
    :toctree: generated/
    :nosignatures:

    RobotJoints
    ExternalAxes

Debugging instructions
----------------------

Wrapping any instruction in a :class:`~compas_rrc.Debug` allows to get raw access to the
output values::


    # Get joints
    raw_debug_output = abb.send_and_wait(rrc.Debug(rrc.GetJoints()))

    # Print received values
    print(raw_debug_output)

.. autosummary::
    :toctree: generated/
    :nosignatures:

    Debug

"""

from __future__ import absolute_import

from compas_rrc.__version__ import (
    __author__,
    __author_email__,
    __copyright__,
    __license__,
    __url__,
    __version__
)
from compas_rrc.client import (
    AbbClient,
    RosClient
)
from compas_rrc.common import (
    CLIENT_PROTOCOL_VERSION,
    ExecutionLevel,
    ExternalAxes,
    FeedbackLevel,
    FutureResult,
    InstructionException,
    RobotJoints,
    TimeoutException
)
from compas_rrc.custom import CustomInstruction
from compas_rrc.io import (
    PulseDigital,
    ReadAnalog,
    ReadDigital,
    ReadGroup,
    SetAnalog,
    SetDigital,
    SetGroup
)
from compas_rrc.motion import (
    Motion,
    MoveToFrame,
    MoveToJoints,
    MoveToRobtarget,
    Zone
)
from compas_rrc.msg import PrintText
from compas_rrc.utility import (
    Debug,
    GetFrame,
    GetJoints,
    GetRobtarget,
    Noop,
    SetAcceleration,
    SetMaxSpeed,
    SetTool,
    SetWorkObject,
    Stop,
    WaitTime
)
from compas_rrc.watch import (
    ReadWatch,
    StartWatch,
    StopWatch
)

__all_plugins__ = ['compas_rrc.__install']
__all__ = [
    '__url__',
    '__version__',
    '__author__',
    '__author_email__',
    '__license__',
    '__copyright__',
    'CLIENT_PROTOCOL_VERSION',
    'FeedbackLevel',
    'ExecutionLevel',
    'InstructionException',
    'TimeoutException',
    'FutureResult',
    'ExternalAxes',
    'RobotJoints',
    'RosClient',
    'AbbClient',
    'SetDigital',
    'SetAnalog',
    'SetGroup',
    'PulseDigital',
    'ReadAnalog',
    'ReadDigital',
    'ReadGroup',
    'Zone',
    'Motion',
    'MoveToJoints',
    'MoveToFrame',
    'MoveToRobtarget',
    'PrintText',
    'CustomInstruction',
    'Noop',
    'GetFrame',
    'GetJoints',
    'GetRobtarget',
    'SetAcceleration',
    'SetTool',
    'SetMaxSpeed',
    'Stop',
    'WaitTime',
    'SetWorkObject',
    'Debug',
    'ReadWatch',
    'StartWatch',
    'StopWatch',
]



from compas.geometry import Frame
from compas_fab.backends.ros.messages import ROSmsg

from compas_rrc.common import ExecutionLevel
from compas_rrc.common import ExternalAxes
from compas_rrc.common import FeedbackLevel
from compas_rrc.common import RobotJoints

INSTRUCTION_PREFIX = 'r_RRC_'

__all__ = ['Noop',
           'GetFrame',
           'GetJoints',
           'GetRobtarget',
           'SetAcceleration',
           'SetTool',
           'SetMaxSpeed',
           'Stop',
           'WaitTime',
           'SetWorkObject',
           'Debug']


def is_rapid_none(val):
    """In RAPID, None values are expressed as 9E+9, they end up as 8999999488 in Python"""
    return int(val) == 8999999488


class Noop(ROSmsg):
    """No-op is a call without any effect. But like all other instructions it makes a roundtrip from the user code to the robot and back.

    Examples
    --------
    .. code-block:: python

        # Noop
        done = abb.send_and_wait(rrc.Noop())

    RAPID Instruction: ``none``

    .. include:: ../abb-reference.rst

    """

    def __init__(self, feedback_level=FeedbackLevel.NONE):
        """Create a new instance of the instruction.

        Parameters
        ----------
        feedback_level : :obj:`int`
            Defines the feedback level requested from the robot. Defaults to :attr:`FeedbackLevel.NONE`.
        """
        self.instruction = INSTRUCTION_PREFIX + 'Noop'
        self.feedback_level = feedback_level
        self.exec_level = ExecutionLevel.ROBOT
        self.string_values = []
        self.float_values = []


class Debug(ROSmsg):
    """Activate debug mode on any instruction by wrapping it.

    Examples
    --------
    .. code-block:: python

        # Get joints
        raw_debug_output = abb.send_and_wait(rrc.Debug(rrc.GetJoints()))

        # Print received values
        print(raw_debug_output)

    """

    def __init__(self, instruction, debug_parser=None):
        """Initialize a new debug instruction wrapping another instruction.

        Parameters
        ----------
        instruction : :class:`ROSmsg`
            Any instruction inheriting from ROS message.
        debug_parser : callable
            Function to be used for parsing the feedback. Optional.
        """
        self._instruction = instruction
        self.debug_parser = debug_parser

    @property
    def msg(self):
        """Raw message."""
        return self._instruction.msg

    @property
    def instruction(self):
        """Name of the instruction."""
        return self._instruction.instruction

    @property
    def sequence_id(self):
        """Sequence identifier."""
        return self._instruction.sequence_id

    @sequence_id.setter
    def sequence_id(self, value):
        self._instruction.sequence_id = value

    @property
    def feedback_level(self):
        """Feedback level."""
        return self._instruction.feedback_level

    @feedback_level.setter
    def feedback_level(self, value):
        self._instruction.feedback_level = value

    @property
    def exec_level(self):
        """Execution level."""
        return self._instruction.exec_level

    @property
    def string_values(self):
        """List of string values."""
        return self._instruction.string_values

    @property
    def float_values(self):
        """List of float values."""
        return self._instruction.float_values

    def parse_feedback(self, result):
        if self.debug_parser:
            return self.debug_parser(result)
        return result


class GetJoints(ROSmsg):
    """Get joints is a call that requests the joint values (:class:`RobotJoints`) and the external axes (:class:`ExternalAxes`) of the robot.

    Examples
    --------
    .. code-block:: python

        # Get joints
        robot_joints, external_axes = abb.send_and_wait(rrc.GetJoints())

    RAPID Instruction: ``CJointT``

    .. include:: ../abb-reference.rst

    """

    def __init__(self):
        """Create a new instance of the instruction."""
        self.instruction = INSTRUCTION_PREFIX + 'GetJoints'
        self.feedback_level = FeedbackLevel.DONE
        self.exec_level = ExecutionLevel.ROBOT
        self.string_values = []
        self.float_values = []

    def parse_feedback(self, result):
        """Parses the result as :class:`RobotJoints` and :class:`ExternalAxes`.

        Return
        ------
        :class:`RobotJoints`, :class:`ExternalAxes`
            Current joints and external axes of the robot.
        """
        # read robot jonts
        robot_joints = [result['float_values'][i] for i in range(0, 6)]

        # read external axes
        external_axes = [result['float_values'][i] for i in range(6, 12) if not is_rapid_none(result['float_values'][i])]

        # write result
        return RobotJoints(*robot_joints), ExternalAxes(*external_axes)


class GetRobtarget(ROSmsg):
    """Instruction to request the current robtarget defined as frame (:class:`compas.geometry.Frame`)
    and external axes (:class:`ExternalAxes`) of the robot.

    Examples
    --------
    .. code-block:: python

        # Get frame and external axes
        frame, external_axes = abb.send_and_wait(rrc.GetRobtarget())

    RAPID Instruction: ``CRobT``

    .. include:: ../abb-reference.rst

    """

    def __init__(self):
        """Create a new instance of the instruction."""
        self.instruction = INSTRUCTION_PREFIX + 'GetRobtarget'
        self.feedback_level = FeedbackLevel.DONE
        self.exec_level = ExecutionLevel.ROBOT
        self.string_values = []
        self.float_values = []

    def parse_feedback(self, result):
        """Parses the result as a :class:`compas.geometry.Frame` and :class:`ExternalAxes`.

        Return
        ------
        :class:`compas.geometry.Frame`, :class:`ExternalAxes`
            Current frame and external axes of the robot.
        """

        # read pos
        x = result['float_values'][0]
        y = result['float_values'][1]
        z = result['float_values'][2]
        pos = [x, y, z]

        # read orient
        orient_q1 = result['float_values'][3]
        orient_q2 = result['float_values'][4]
        orient_q3 = result['float_values'][5]
        orient_q4 = result['float_values'][6]
        orientation = [orient_q1, orient_q2, orient_q3, orient_q4]

        # read gantry joints
        external_axes = [result['float_values'][i] for i in range(7, 13) if not is_rapid_none(result['float_values'][i])]

        # write result

        # As compas frame
        result = Frame.from_quaternion(orientation, point=pos)

        # End
        return result, ExternalAxes(*external_axes)


class GetFrame(GetRobtarget):
    """Instruction to request the current frame (:class:`compas.geometry.Frame`) of the robot.

    Examples
    --------
    .. code-block:: python

        # Get frame
        frame = abb.send_and_wait(rrc.GetFrame())

    RAPID Instruction: ``CRobT``

    .. include:: ../abb-reference.rst

    """

    def parse_feedback(self, result):
        """Parses the result as a :class:`compas.geometry.Frame`.

        Return
        ------
        :class:`compas.geometry.Frame`
            Current frame of the robot.
        """
        frame, _ext_axes = super(GetFrame, self).parse_feedback(result)
        return frame


class SetAcceleration(ROSmsg):
    """Set acceleration is a call that sets the acc- and deceleration of the robot.

    Examples
    --------
    .. code-block:: python

        # Set acceleration
        acc = 100 # Unit [%]
        ramp = 100  # Unit [%]
        done = abb.send_and_wait(rrc.SetAcceleration(acc, ramp))

    RAPID Instruction: ``AccSet``

    .. include:: ../abb-reference.rst
    """

    def __init__(self, acc, ramp, feedback_level=FeedbackLevel.NONE):
        """Create a new instance of the instruction.

        Parameters
        ----------
        acc : :obj:`float`
            Acceleration or deceleration expressed in percentage of the system parameters of the robot.
        ramp : :obj:`float`
            The rate at which acceleration or deceleration changes expressed in percentage.
        feedback_level : :obj:`int`
            Defines the feedback level requested from the robot. Defaults to :attr:`FeedbackLevel.NONE`.
        """
        self.instruction = INSTRUCTION_PREFIX + 'SetAcceleration'
        self.feedback_level = feedback_level
        self.exec_level = ExecutionLevel.ROBOT
        self.string_values = []
        self.float_values = [acc, ramp]


class SetTool(ROSmsg):
    """Set tool is a call that sets a predefined tool in the robot as active.

    Examples
    --------
    .. code-block:: python

        # Set tool
        done = abb.send_and_wait(rrc.SetTool('tool0'))

    RAPID Instruction: ``tooldata``

    .. include:: ../abb-reference.rst

    """

    def __init__(self, tool_name, feedback_level=FeedbackLevel.NONE):
        """Create a new instance of the instruction.

        Parameters
        ----------
        tool_name : :obj:`str`
            Name of the tool as defined in RAPID.
        feedback_level : :obj:`int`
            Defines the feedback level requested from the robot. Defaults to :attr:`FeedbackLevel.NONE`.
        """
        self.instruction = INSTRUCTION_PREFIX + 'SetTool'
        self.feedback_level = feedback_level
        self.exec_level = ExecutionLevel.ROBOT
        self.string_values = [tool_name]
        self.float_values = []


class SetMaxSpeed(ROSmsg):
    """Set max speed is a call that sets the override and maximal tool center point (TCP) speed from the robot.

    Examples
    --------
    .. code-block:: python

        # Set max speed
        override = 100 # Unit [%]
        max_tcp = 2500 # Unit [mm/s]
        done = abb.send_and_wait(rrc.SetMaxSpeed(override, max_tcp))

    RAPID Instruction: ``VelSet``

    .. include:: ../abb-reference.rst

    """

    def __init__(self, override, max_tcp, feedback_level=FeedbackLevel.NONE):
        """Create a new instance of the instruction.

        Parameters
        ----------
        override : :obj:`float`
            Internal speed override expressed in percentage. This override is additional to the one that
            can be defined on the robot panel. It is recommended to use only one of them, preferably the
            one on the robot panel, and leave this to 100%.
        max_tcp : :obj:`float`
            Maximum tool center point (TCP) speed expressed in mm/sec. All instructions following a ``SetMaxSpeed``
            are affected by it.
        feedback_level : :obj:`int`
            Defines the feedback level requested from the robot. Defaults to :attr:`FeedbackLevel.NONE`.
        """
        self.instruction = INSTRUCTION_PREFIX + 'SetMaxSpeed'
        self.feedback_level = feedback_level
        self.exec_level = ExecutionLevel.ROBOT
        self.string_values = []
        self.float_values = [override, max_tcp]


class SetWorkObject(ROSmsg):
    """Set work object is a call that sets a predefined work object in the robot as active.

    Examples
    --------
    .. code-block:: python

        # Set work object
        done = abb.send_and_wait(rrc.SetWorkObject('wobj0'))

    RAPID Instruction: ``wobjdata``

    .. include:: ../abb-reference.rst

    """

    def __init__(self, wobj_name, feedback_level=FeedbackLevel.NONE):
        """Create a new instance of the instruction.

        Parameters
        ----------
        wobj_name : :obj:`str`
            Name of the work object as defined in RAPID.
        feedback_level : :obj:`int`
            Defines the feedback level requested from the robot. Defaults to :attr:`FeedbackLevel.NONE`.
        """
        self.instruction = INSTRUCTION_PREFIX + 'SetWorkObject'
        self.feedback_level = feedback_level
        self.exec_level = ExecutionLevel.ROBOT
        self.string_values = [wobj_name]
        self.float_values = []


class Stop(ROSmsg):
    """Stop is a function to stop the associated motion task of the robot.

    Examples
    --------
    .. code-block:: python

        # Stop
        done = abb.send_and_wait(rrc.Stop())

    RAPID Instruction: ``Stop``

    .. include:: ../abb-reference.rst

    """

    def __init__(self, feedback_level=FeedbackLevel.NONE):
        """Create a new instance of the instruction.

        Parameters
        ----------
        feedback_level : :obj:`int`
            Defines the feedback level requested from the robot. Defaults to :attr:`FeedbackLevel.NONE`.
        """
        self.instruction = INSTRUCTION_PREFIX + 'Stop'
        self.feedback_level = feedback_level
        self.exec_level = ExecutionLevel.ROBOT
        self.string_values = []
        self.float_values = []


class WaitTime(ROSmsg):
    """Wait time is a call that will wait on the robot task for a certain time.

    Examples
    --------
    .. code-block:: python

        # Wait time
        time = 1.0 # Unit [s]
        done = abb.send_and_wait(rrc.WaitTime(time))

    RAPID Instruction: ``WaitTime``

    .. include:: ../abb-reference.rst

    """

    def __init__(self, time, feedback_level=FeedbackLevel.NONE):
        """Create a new instance of the instruction.

        Parameters
        ----------
        time : :obj:`float`
            Duration of the wait expressed in seconds.
        feedback_level : :obj:`int`
            Defines the feedback level requested from the robot. Defaults to :attr:`FeedbackLevel.NONE`.
        """
        self.instruction = INSTRUCTION_PREFIX + 'WaitTime'
        self.feedback_level = feedback_level
        self.exec_level = ExecutionLevel.ROBOT
        self.string_values = []
        self.float_values = [time]
